from __future__ import unicode_literals

import unittest

from fs.errors import FSError
from fs.memoryfs import MemoryFS
from fs import walk


class TestWalkerBase(unittest.TestCase):

    def test_not_implemented(self):
        walker = walk.WalkerBase()
        with self.assertRaises(NotImplementedError):
            walker.walk(None, path='/')


class TestWalker(unittest.TestCase):

    def setUp(self):
        self.walker = walk.Walker()

    def test_repr(self):
        repr(self.walker)

    def test_create(self):
        with self.assertRaises(AssertionError):
            walk.Walker(ignore_errors=True, on_error=lambda path, error: True)
        walk.Walker(ignore_errors=True)


class TestWalk(unittest.TestCase):

    def setUp(self):
        self.fs = MemoryFS()

        self.fs.makedir('foo1')
        self.fs.makedir('foo2')
        self.fs.makedir('foo3')
        self.fs.create('foo1/top1.txt')
        self.fs.create('foo1/top2.txt')
        self.fs.makedir('foo1/bar1')
        self.fs.makedir('foo2/bar2')
        self.fs.makedir('foo2/bar2/bar3')
        self.fs.create('foo2/bar2/bar3/test.txt')
        self.fs.create('foo2/top3.txt')

    def test_invalid(self):
        with self.assertRaises(ValueError):
            self.fs.walk(search='random')

    def test_repr(self):
        repr(self.fs.walk)

    def test_walk(self):
        walk = []
        for path, dirs, files in self.fs.walk():
            walk.append((
                path,
                [info.name for info in dirs],
                [info.name for info in files]
            ))
        expected = [(u'/', [u'foo1', u'foo2', u'foo3'], []), (u'/foo1', [u'bar1'], [u'top1.txt', u'top2.txt']), (u'/foo2', [u'bar2'], [u'top3.txt']), (u'/foo3', [], []), (u'/foo1/bar1', [], []), (u'/foo2/bar2', [u'bar3'], []), (u'/foo2/bar2/bar3', [], [u'test.txt'])]
        self.assertEqual(walk, expected)

    def test_walk_files(self):
        files = list(self.fs.walk.files())

        self.assertEqual(
            files,
            [
                '/foo1/top1.txt',
                '/foo1/top2.txt',
                '/foo2/top3.txt',
                '/foo2/bar2/bar3/test.txt',
            ]
        )

        files = list(self.fs.walk.files(search="depth"))
        self.assertEqual(
            files,
            [
                '/foo1/top1.txt',
                '/foo1/top2.txt',
                '/foo2/bar2/bar3/test.txt',
                '/foo2/top3.txt',
            ]
        )

    def test_walk_dirs(self):
        dirs = list(self.fs.walk.dirs())
        self.assertEqual(
            dirs,
            [
                '/foo1',
                '/foo2',
                '/foo3',
                '/foo1/bar1',
                '/foo2/bar2',
                '/foo2/bar2/bar3'
            ]
        )

        dirs = list(self.fs.walk.dirs(search="depth"))
        self.assertEqual(
            dirs,
            [
                '/foo1/bar1',
                '/foo2/bar2/bar3',
                '/foo2/bar2',
                '/foo1',
                '/foo2',
                '/foo3'
            ]
        )

        dirs = list(self.fs.walk.dirs(search="depth", exclude_dirs=['foo2']))
        self.assertEqual(
            dirs,
            [
                '/foo1/bar1',
                '/foo1',
                '/foo3'
            ]
        )

    def test_walk_info(self):
        walk = []
        for path, info in self.fs.walk.info():
            walk.append((path, info.is_dir, info.name))
        expected = [(u'/foo1', True, u'foo1'), (u'/foo2', True, u'foo2'), (u'/foo3', True, u'foo3'), (u'/foo1/bar1', True, u'bar1'), (u'/foo1/top1.txt', False, u'top1.txt'), (u'/foo1/top2.txt', False, u'top2.txt'), (u'/foo2/bar2', True, u'bar2'), (u'/foo2/top3.txt', False, u'top3.txt'), (u'/foo2/bar2/bar3', True, u'bar3'), (u'/foo2/bar2/bar3/test.txt', False, u'test.txt')]
        self.assertEqual(walk, expected)

    def test_broken(self):
        original_scandir = self.fs.scandir

        def broken_scandir(path, namespaces=None):
            if path == '/foo2':
                raise FSError("can't read dir")
            return original_scandir(path, namespaces=namespaces)

        self.fs.scandir = broken_scandir

        files = list(self.fs.walk.files(search="depth"))
        self.assertEqual(
            files,
            [
                '/foo1/top1.txt',
                '/foo1/top2.txt'
            ]
        )

        with self.assertRaises(FSError):
            list(
                self.fs.walk.files(
                    on_error=lambda path, error: False
                )
            )



from __future__ import unicode_literals

import errno
import datetime
import os
import unittest
import tempfile
import shutil
import datetime
from six import PY2

import fs.copy
from fs import open_fs


class TestCopy(unittest.TestCase):

    def test_copy_fs(self):
        src_fs = open_fs('mem://')
        src_fs.makedirs('foo/bar')
        src_fs.makedirs('foo/empty')
        src_fs.touch('test.txt')
        src_fs.touch('foo/bar/baz.txt')

        dst_fs = open_fs('mem://')
        fs.copy.copy_fs(src_fs, dst_fs)

        self.assertTrue(dst_fs.isdir('foo/empty'))
        self.assertTrue(dst_fs.isdir('foo/bar'))
        self.assertTrue(dst_fs.isfile('test.txt'))

    def test_copy_dir(self):
        src_fs = open_fs('mem://')
        src_fs.makedirs('foo/bar')
        src_fs.makedirs('foo/empty')
        src_fs.touch('test.txt')
        src_fs.touch('foo/bar/baz.txt')

        dst_fs = open_fs('mem://')
        fs.copy.copy_dir(src_fs, '/foo', dst_fs, '/')

        self.assertTrue(dst_fs.isdir('bar'))
        self.assertTrue(dst_fs.isdir('empty'))
        self.assertTrue(dst_fs.isfile('bar/baz.txt'))

    def test_copy_dir_on_copy(self):
        src_fs = open_fs('mem://')
        src_fs.touch('baz.txt')

        on_copy_calls = []
        def on_copy(*args):
            on_copy_calls.append(args)

        dst_fs = open_fs('mem://')
        fs.copy.copy_dir(src_fs, '/', dst_fs, '/', on_copy=on_copy)
        print(on_copy_calls)
        self.assertEqual(
            on_copy_calls,
            [(src_fs, '/baz.txt', dst_fs, '/baz.txt')]
        )

    def mkdirp(self, path):
        #os.makedirs(path, exist_ok=True) only for python3.?
        try:
            os.makedirs(path)
        except OSError as exc:
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise

    def _create_sandbox_dir(self, prefix='pyfilesystem2_sandbox_', home=None):
        if home is None:
            return tempfile.mkdtemp(prefix=prefix)
        else:
            sandbox_path = os.path.join(home, prefix)
            self.mkdirp(sandbox_path)
            return sandbox_path

    def _touch(self, root, filepath):
        #create abs filename
        abs_filepath = os.path.join(root, filepath)
        #create dir
        dirname = os.path.dirname(abs_filepath)
        self.mkdirp(dirname)
        #touch file
        with open(abs_filepath, 'a'):
            os.utime(abs_filepath, None) #update the mtime in case the file exists, same as touch

        return abs_filepath

    def _write_file(self, filepath, write_chars=1024):
        with open(filepath, 'w') as f:
            f.write('1' * write_chars)
        return filepath

    def _delay_file_utime(self, filepath, delta_sec=None):
        import calendar
        from datetime import datetime
        file_access_mod_time = int(calendar.timegm(datetime.now().timetuple())) + delta_sec
        times = (file_access_mod_time, file_access_mod_time)
        os.utime(filepath, times)

    def test_copy_file_if_newer_same_fs(self):
        src_fs = open_fs('mem://')
        src_fs.makedir('foo2').touch('exists')
        src_fs.makedir('foo1').touch('test1.txt')
        src_fs.settimes('foo2/exists', datetime.datetime.utcnow() + datetime.timedelta(hours=1))
        self.assertTrue(fs.copy.copy_file_if_newer(src_fs, 'foo1/test1.txt', src_fs, 'foo2/test1.txt.copy'))
        self.assertFalse(fs.copy.copy_file_if_newer(src_fs, 'foo1/test1.txt', src_fs, 'foo2/exists'))
        self.assertTrue(src_fs.exists('foo2/test1.txt.copy'))

    def test_copy_file_if_newer_dst_older(self):
        try:
            #create first dst ==> dst is older the src ==> file should be copied
            dst_dir = self._create_sandbox_dir()
            dst_file1 = self._touch(dst_dir, "file1.txt")
            self._write_file(dst_file1)

            src_dir = self._create_sandbox_dir()
            src_file1 = self._touch(src_dir, "file1.txt")
            self._write_file(src_file1)
            #ensure src file is newer than dst, changing its modification time
            self._delay_file_utime(src_file1, delta_sec=60)

            src_fs = open_fs('osfs://' + src_dir)
            dst_fs = open_fs('osfs://' + dst_dir)

            self.assertTrue(dst_fs.exists("/file1.txt"))

            copied = fs.copy.copy_file_if_newer(src_fs, "/file1.txt", dst_fs, "/file1.txt")

            self.assertTrue(copied)
            self.assertTrue(dst_fs.exists("/file1.txt"))
        finally:
            shutil.rmtree(src_dir)
            shutil.rmtree(dst_dir)

    def test_copy_file_if_newer_dst_doesnt_exists(self):
        try:
            src_dir = self._create_sandbox_dir()
            src_file1 = self._touch(src_dir, "file1.txt")
            self._write_file(src_file1)

            dst_dir = self._create_sandbox_dir()

            src_fs = open_fs('osfs://' + src_dir)
            dst_fs = open_fs('osfs://' + dst_dir)


            copied = fs.copy.copy_file_if_newer(src_fs, "/file1.txt", dst_fs, "/file1.txt")

            self.assertTrue(copied)
            self.assertTrue(dst_fs.exists("/file1.txt"))
        finally:
            shutil.rmtree(src_dir)
            shutil.rmtree(dst_dir)

    def test_copy_file_if_newer_dst_is_newer(self):
        try:
            src_dir = self._create_sandbox_dir()
            src_file1 = self._touch(src_dir, "file1.txt")
            self._write_file(src_file1)

            dst_dir = self._create_sandbox_dir()
            dst_file1 = self._touch(dst_dir, "file1.txt")
            self._write_file(dst_file1)

            src_fs = open_fs('osfs://' + src_dir)
            dst_fs = open_fs('osfs://' + dst_dir)


            self.assertTrue(dst_fs.exists("/file1.txt"))

            copied = fs.copy.copy_file_if_newer(src_fs, "/file1.txt", dst_fs, "/file1.txt")

            self.assertEqual(copied, False)
        finally:
            shutil.rmtree(src_dir)
            shutil.rmtree(dst_dir)


    def test_copy_fs_if_newer_dst_older(self):
        try:
            #create first dst ==> dst is older the src ==> file should be copied
            dst_dir = self._create_sandbox_dir()
            dst_file1 = self._touch(dst_dir, "file1.txt")
            self._write_file(dst_file1)

            src_dir = self._create_sandbox_dir()
            src_file1 = self._touch(src_dir, "file1.txt")
            self._write_file(src_file1)
            #ensure src file is newer than dst, changing its modification time
            self._delay_file_utime(src_file1, delta_sec=60)

            src_fs = open_fs('osfs://' + src_dir)
            dst_fs = open_fs('osfs://' + dst_dir)

            self.assertTrue(dst_fs.exists("/file1.txt"))

            copied = []
            def on_copy(src_fs, src_path, dst_fs, dst_path):
                copied.append(dst_path)

            fs.copy.copy_fs_if_newer(src_fs, dst_fs, on_copy=on_copy)

            self.assertEqual(copied, ["/file1.txt"])
            self.assertTrue(dst_fs.exists("/file1.txt"))

            src_fs.close()
            dst_fs.close()

        finally:
            shutil.rmtree(src_dir)
            shutil.rmtree(dst_dir)

    def test_copy_fs_if_newer_when_dst_doesnt_exists(self):
        try:
            src_dir = self._create_sandbox_dir()
            src_file1 = self._touch(src_dir, "file1.txt")
            self._write_file(src_file1)

            src_file2 = self._touch(src_dir, "one_level_down" + os.sep + "file2.txt")
            self._write_file(src_file2)

            dst_dir = self._create_sandbox_dir()

            src_fs = open_fs('osfs://' + src_dir)
            dst_fs = open_fs('osfs://' + dst_dir)

            copied = []
            def on_copy(src_fs, src_path, dst_fs, dst_path):
                copied.append(dst_path)

            fs.copy.copy_fs_if_newer(src_fs, dst_fs, on_copy=on_copy)

            self.assertEqual(copied, ["/file1.txt", "/one_level_down/file2.txt"])
            self.assertTrue(dst_fs.exists("/file1.txt"))
            self.assertTrue(dst_fs.exists("/one_level_down/file2.txt"))

            src_fs.close()
            dst_fs.close()

        finally:
            shutil.rmtree(src_dir)
            shutil.rmtree(dst_dir)

    def test_copy_fs_if_newer_dont_copy_when_dst_exists(self):
        try:
            #src is older than dst => no copy should be necessary
            src_dir = self._create_sandbox_dir()
            src_file1 = self._touch(src_dir, "file1.txt")
            self._write_file(src_file1)

            dst_dir = self._create_sandbox_dir()
            dst_file1 = self._touch(dst_dir, "file1.txt")
            self._write_file(dst_file1)
            #ensure dst file is newer than src, changing its modification time
            self._delay_file_utime(dst_file1, delta_sec=60)

            src_fs = open_fs('osfs://' + src_dir)
            dst_fs = open_fs('osfs://' + dst_dir)

            self.assertTrue(dst_fs.exists("/file1.txt"))

            copied = []
            def on_copy(src_fs, src_path, dst_fs, dst_path):
                copied.append(dst_path)

            fs.copy.copy_fs_if_newer(src_fs, dst_fs, on_copy=on_copy)

            self.assertEqual(copied, [])
            self.assertTrue(dst_fs.exists("/file1.txt"))

            src_fs.close()
            dst_fs.close()

        finally:
            shutil.rmtree(src_dir)
            shutil.rmtree(dst_dir)

    def test_copy_dir_if_newer_one_dst_doesnt_exist(self):
        try:

            src_dir = self._create_sandbox_dir()
            src_file1 = self._touch(src_dir, "file1.txt")
            self._write_file(src_file1)

            src_file2 = self._touch(src_dir, "one_level_down" + os.sep + "file2.txt")
            self._write_file(src_file2)

            dst_dir = self._create_sandbox_dir()
            dst_file1 = self._touch(dst_dir, "file1.txt")
            self._write_file(dst_file1)
            #ensure dst file is newer than src, changing its modification time
            self._delay_file_utime(dst_file1, delta_sec=60)

            src_fs = open_fs('osfs://' + src_dir)
            dst_fs = open_fs('osfs://' + dst_dir)

            copied = []
            def on_copy(src_fs, src_path, dst_fs, dst_path):
                copied.append(dst_path)

            fs.copy.copy_dir_if_newer(src_fs, "/", dst_fs, "/", on_copy=on_copy)

            self.assertEqual(copied, ["/one_level_down/file2.txt"])
            self.assertTrue(dst_fs.exists("/one_level_down/file2.txt"))

            src_fs.close()
            dst_fs.close()
        finally:
            shutil.rmtree(src_dir)
            shutil.rmtree(dst_dir)

    def test_copy_dir_if_newer_same_fs(self):
        try:
            src_dir = self._create_sandbox_dir()
            src_file1 = self._touch(src_dir, "src" + os.sep + "file1.txt")
            self._write_file(src_file1)

            dst_dir = self._create_sandbox_dir(home=src_dir)

            src_fs = open_fs('osfs://' + src_dir)

            copied = []
            def on_copy(src_fs, src_path, dst_fs, dst_path):
                copied.append(dst_path)

            fs.copy.copy_dir_if_newer(src_fs, "/src", src_fs, "/dst", on_copy=on_copy)

            self.assertEqual(copied, ["/dst/file1.txt"])
            self.assertTrue(src_fs.exists("/dst/file1.txt"))

            src_fs.close()

        finally:
            shutil.rmtree(src_dir)

    def test_copy_dir_if_newer_multiple_files(self):

        try:
            src_dir = self._create_sandbox_dir()
            src_fs = open_fs('osfs://' + src_dir)
            src_fs.makedirs('foo/bar')
            src_fs.makedirs('foo/empty')
            src_fs.touch('test.txt')
            src_fs.touch('foo/bar/baz.txt')

            dst_dir = self._create_sandbox_dir()
            dst_fs = open_fs('osfs://' + dst_dir)

            fs.copy.copy_dir_if_newer(src_fs, '/foo', dst_fs, '/')

            self.assertTrue(dst_fs.isdir('bar'))
            self.assertTrue(dst_fs.isdir('empty'))
            self.assertTrue(dst_fs.isfile('bar/baz.txt'))
        finally:
            shutil.rmtree(src_dir)
            shutil.rmtree(dst_dir)

if __name__ == "__main__":
    unittest.main()


from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from copy import deepcopy

from .path import join
from .enums import ResourceType
from .errors import MissingInfoNamespace
from .permissions import Permissions
from .time import epoch_to_datetime


class Info(object):
    """
    Container for :ref:`info`, returned by the following methods:

     * :meth:`~fs.base.FS.getinfo`
     * :meth:`~fs.base.FS.scandir`
     * :meth:`~fs.base.FS.filterfir`

    :param dict raw_info: A dict containing resource info.
    :param to_datetime: A callable that converts an epoch time to a
        datetime object. The default uses
        :func:`~fs.time.epoch_to_datetime`.

    """

    def __init__(self,
                 raw_info,
                 to_datetime=epoch_to_datetime):
        """
        Create a resource info object from a raw info dict.

        """
        self.raw = raw_info
        self._to_datetime = to_datetime
        self.namespaces = frozenset(self.raw.keys())

    def __repr__(self):
        if self.is_dir:
            return "<dir '{}'>".format(self.name)
        else:
            return "<file '{}'>".format(self.name)

    def __eq__(self, other):
        return self.raw == other.raw

    def _make_datetime(self, t):
        if t is not None:
            return self._to_datetime(t)
        else:
            return None

    def get(self, namespace, key, default=None):
        """
        Get a raw info value.

        >>> info.get('access', 'permissions')
        ['u_r', 'u_w', '_wx']

        :param str namespace: A namespace identifier.
        :param str key: A key within the namespace.
        :param default: A default value to return if either the
            namespace or namespace + key is not found.
        """
        try:
            return self.raw[namespace].get(key, default)
        except KeyError:
            return default

    def _require_namespace(self, namespace):
        """
        Raise a MissingInfoNamespace if the given namespace is not
        present in the info.

        """
        if namespace not in self.raw:
            raise MissingInfoNamespace(namespace)

    def is_writeable(self, namespace, key):
        """
        Check if a given key in a namespace is writable (with
        :meth:`~fs.base.FS.setinfo`).

        :param namespace: A namespace identifier.
        :type namespace: str
        :param key: A key within the namespace.
        :type key: str
        :rtype: bool

        """
        _writeable = self.get(namespace, '_write', ())
        return key in _writeable

    def has_namespace(self, namespace):
        """
        Check if the resource info contains a given namespace.

        :param namespace: A namespace name.
        :type namespace: str
        :rtype: bool

        """
        return namespace in self.raw

    def copy(self, to_datetime=None):
        """Create a copy of this resource info object."""
        return Info(
            deepcopy(self.raw),
            to_datetime=to_datetime or self._to_datetime
        )

    def make_path(self, dir_path):
        """
        Make a path by joining ``dir_path`` with the resource name.

        :param dir_path: A path to a directory.
        :type dir_path: str
        :returns: A path.
        :rtype: str

        """
        return join(dir_path, self.name)

    @property
    def name(self):
        """
        Get the resource name.

        :rtype: str

        """
        return self.get('basic', 'name')

    @property
    def is_dir(self):
        """
        Check if the resource references a directory.

        :rtype: bool

        """
        return self.get('basic', 'is_dir')

    @property
    def is_file(self):
        """
        Check if a resource references a file.

        :rtype: bool

        """
        return not self.get('basic', 'is_dir')

    @property
    def is_link(self):
        """
        Check if a resource is a symlink.

        :rtype: bool

        """
        self._require_namespace('link')
        return self.get('link', 'target') is not None

    @property
    def type(self):
        """
        Get the resource type enumeration.

        Requires the ``"details"`` namespace.

        :type: :class:`~fs.ResourceType`
        :raises ~fs.errors.MissingInfoNamespace: if the 'details'
            namespace is not in the Info.

        """
        self._require_namespace('details')
        return ResourceType(self.get('details', 'type', 0))

    @property
    def accessed(self):
        """
        Get the time this resource was last accessed, or ``None`` if not
        available.

        Requires the ``"details"`` namespace.

        :rtype: datetime
        :raises ~fs.errors.MissingInfoNamespace: if the 'details'
            namespace is not in the Info.

        """
        self._require_namespace('details')
        _time = self._make_datetime(
            self.get('details', 'accessed')
        )
        return _time

    @property
    def modified(self):
        """
        Get the time the resource was modified, or ``None`` if not
        available.

        Requires the ``"details"`` namespace.

        :rtype: datetime
        :raises ~fs.errors.MissingInfoNamespace: if the 'details'
            namespace is not in the Info.

        """
        self._require_namespace('details')
        _time = self._make_datetime(
            self.get('details', 'modified')
        )
        return _time

    @property
    def created(self):
        """
        Get the time this resource was created, or ``None`` if not
        available.

        Requires the ``"details"`` namespace.

        :rtype: datetime
        :raises ~fs.errors.MissingInfoNamespace: if the 'details'
            namespace is not in the Info.

        """
        self._require_namespace('details')
        _time = self._make_datetime(
            self.get('details', 'created')
        )
        return _time

    @property
    def metadata_changed(self):
        """
        Get the time the metadata changed, or ``None`` if not
        available.

        Requires the ``"details"`` namespace.

        :rtype: datetime
        :raises ~fs.errors.MissingInfoNamespace: if the 'details'
            namespace is not in the Info.

        """
        self._require_namespace('details')
        _time = self._make_datetime(
            self.get('details', 'metadata_changed')
        )
        return _time

    @property
    def permissions(self):
        """
        Get a permissions object, or ``None`` if not available.

        Requires the ``"access"`` namespace.

        :rtype: :class:`fs.permissions.Permissions`
        :raises ~fs.errors.MissingInfoNamespace: if the 'access'
            namespace is not in the Info.

        """
        self._require_namespace('access')
        _perm_names = self.get('access', 'permissions')
        if _perm_names is None:
            return None
        permissions = Permissions(_perm_names)
        return permissions

    @property
    def size(self):
        """
        Get the size of the resource, in bytes.

        Requires the ``"details"`` namespace.

        :rtype: int
        :raises ~fs.errors.MissingInfoNamespace: if the 'details'
            namespace is not in the Info.

        """
        self._require_namespace('details')
        return self.get('details', 'size')

    @property
    def user(self):
        """
        Get the owner of a resource, or ``None`` if not available.

        Requires the ``"access"`` namespace.

        :rtype: str
        :raises ~fs.errors.MissingInfoNamespace: if the 'access'
            namespace is not in the Info.

        """
        self._require_namespace('access')
        return self.get('access', 'user')

    @property
    def uid(self):
        """
        Get the user id of a resource, or ``None`` if not available.

        Requires the ``"access"`` namespace.

        :rtype: int
        :raises ~fs.errors.MissingInfoNamespace: if the 'access'
            namespace is not in the Info.

        """
        self._require_namespace('access')
        return self.get('access', 'uid')

    @property
    def group(self):
        """
        Get the group of the resource owner, or ``None`` if not
        available.

        Requires the ``"access"`` namespace.

        :rtype: str
        :raises ~fs.errors.MissingInfoNamespace: if the 'access'
            namespace is not in the Info.

        """
        self._require_namespace('access')
        return self.get('access', 'group')

    @property
    def gid(self):
        """
        Get the group id of a resource, or ``None`` if not available.

        Requires the ``"access"`` namespace.

        :rtype: int
        :raises ~fs.errors.MissingInfoNamespace: if the 'access'
            namespace is not in the Info.

        """
        self._require_namespace('access')
        return self.get('access', 'gid')

    @property
    def target(self):
        """
        Get the link target, if this is a symlink, or ``None`` if this
        is not a symlink.

        Requires the ``"link"`` namespace.

        :rtype: bool
        :raises ~fs.errors.MissingInfoNamespace: if the 'link'
            namespace is not in the Info.

        """
        self._require_namespace('link')
        return self.get('link', 'target')



from __future__ import unicode_literals

import unittest

from fs.multifs import MultiFS
from fs.memoryfs import MemoryFS
from fs import errors

from fs.test import FSTestCases


class TestMultiFS(FSTestCases, unittest.TestCase):
    """Test OSFS implementation."""

    def setUp(self):
        fs = MultiFS()
        mem_fs = MemoryFS()
        fs.add_fs('mem', mem_fs, write=True)
        self.fs = fs
        self.mem_fs = mem_fs

    def make_fs(self):
        fs = MultiFS()
        mem_fs = MemoryFS()
        fs.add_fs('mem', mem_fs, write=True)
        return fs

    def test_get_fs(self):
        self.assertIs(self.fs.get_fs('mem'), self.mem_fs)

    def test_which(self):
        self.fs.setbytes('foo', b'bar')
        self.assertEqual(self.fs.which('foo'), ('mem', self.mem_fs))
        self.assertEqual(self.fs.which('bar', 'w'), ('mem', self.mem_fs))
        self.assertEqual(self.fs.which('baz'), (None, None))

    def test_auto_close(self):
        """Test MultiFS auto close is working"""
        multi_fs = MultiFS()
        m1 = MemoryFS()
        m2 = MemoryFS()
        multi_fs.add_fs('m1', m1)
        multi_fs.add_fs('m2', m2)
        self.assertFalse(m1.isclosed())
        self.assertFalse(m2.isclosed())
        multi_fs.close()
        self.assertTrue(m1.isclosed())
        self.assertTrue(m2.isclosed())

    def test_no_auto_close(self):
        """Test MultiFS auto close can be disabled"""
        multi_fs = MultiFS(auto_close=False)
        self.assertEqual(
            repr(multi_fs),
            "MultiFS(auto_close=False)"
        )
        m1 = MemoryFS()
        m2 = MemoryFS()
        multi_fs.add_fs('m1', m1)
        multi_fs.add_fs('m2', m2)
        self.assertFalse(m1.isclosed())
        self.assertFalse(m2.isclosed())
        multi_fs.close()
        self.assertFalse(m1.isclosed())
        self.assertFalse(m2.isclosed())

    def test_opener(self):
        """Test use of FS URLs."""
        multi_fs = MultiFS()
        with self.assertRaises(TypeError):
            multi_fs.add_fs(u'foo', 5)
        multi_fs.add_fs(u'f1', u'mem://')
        multi_fs.add_fs(u'f2', u'temp://')
        self.assertIsInstance(
            multi_fs.get_fs(u'f1'),
            MemoryFS
        )

    def test_priority(self):
        """Test priority order is working"""
        m1 = MemoryFS()
        m2 = MemoryFS()
        m3 = MemoryFS()
        m1.setbytes("name", b"m1")
        m2.setbytes("name", b"m2")
        m3.setbytes("name", b"m3")
        multi_fs = MultiFS(auto_close=False)
        multi_fs.add_fs("m1", m1)
        multi_fs.add_fs("m2", m2)
        multi_fs.add_fs("m3", m3)
        self.assertEqual(multi_fs.getbytes("name"), b"m3")

        m1 = MemoryFS()
        m2 = MemoryFS()
        m3 = MemoryFS()
        m1.setbytes("name", b"m1")
        m2.setbytes("name", b"m2")
        m3.setbytes("name", b"m3")
        multi_fs = MultiFS(auto_close=False)
        multi_fs.add_fs("m1", m1)
        multi_fs.add_fs("m2", m2, priority=10)
        multi_fs.add_fs("m3", m3)
        self.assertEqual(multi_fs.getbytes("name"), b"m2")

        m1 = MemoryFS()
        m2 = MemoryFS()
        m3 = MemoryFS()
        m1.setbytes("name", b"m1")
        m2.setbytes("name", b"m2")
        m3.setbytes("name", b"m3")
        multi_fs = MultiFS(auto_close=False)
        multi_fs.add_fs("m1", m1)
        multi_fs.add_fs("m2", m2, priority=10)
        multi_fs.add_fs("m3", m3, priority=10)
        self.assertEqual(multi_fs.getbytes("name"), b"m3")

        m1 = MemoryFS()
        m2 = MemoryFS()
        m3 = MemoryFS()
        m1.setbytes("name", b"m1")
        m2.setbytes("name", b"m2")
        m3.setbytes("name", b"m3")
        multi_fs = MultiFS(auto_close=False)
        multi_fs.add_fs("m1", m1, priority=11)
        multi_fs.add_fs("m2", m2, priority=10)
        multi_fs.add_fs("m3", m3, priority=10)
        self.assertEqual(multi_fs.getbytes("name"), b"m1")

    def test_no_writable(self):
        fs = MultiFS()
        with self.assertRaises(errors.ResourceReadOnly):
            fs.setbytes('foo', b'bar')

    def test_validate_path(self):
        self.fs.write_fs = None
        self.fs.validatepath('foo')



from __future__ import print_function
from __future__ import unicode_literals

from datetime import datetime
import zipfile

import six

from . import errors
from .base import FS
from .compress import write_zip
from .enums import ResourceType
from .info import Info
from .iotools import RawWrapper
from .memoryfs import MemoryFS
from .opener import open_fs
from .path import dirname, normpath, relpath
from .time import datetime_to_epoch
from .wrapfs import WrapFS


class ZipFS(WrapFS):
    """
    Read and write zip files.

    There are two ways to open a ZipFS for the use cases of reading
    a zip file, and creating a new one.

    If you open the ZipFS with  ``write`` set to ``False`` (the
    default), then the filesystem will be a read only filesystem which
    maps to the files and directories within the zip file. Files are
    decompressed on the fly when you open them.

    Here's how you might extract and print a readme from a zip file::

        with ZipFS('foo.zip') as zip_fs:
            readme = zip_fs.gettext('readme.txt')

    If you open the ZipFS with ``write`` set to ``True``, then the ZipFS
    will be a empty temporary filesystem. Any files / directories you
    create in the ZipFS will be written in to a zip file when the ZipFS
    is closed.

    Here's how you might write a new zip file containing a readme.txt
    file::

        with ZipFS('foo.zip', write=True) as new_zip:
            new_zip.settext(
                'readme.txt',
                'This zip file was written by PyFilesystem'
            )


    :param file: An OS filename, or a open file object.
    :type file: str or file
    :param write: Set to ``True`` to write a new zip file, or ``False``
        to read an existing zip file.
    :type write: bool
    :param compression:  Compression to use (one of the constants
        defined in the zipfile module in the stdlib).
    :type compression: int
    :param temp_fs: An opener string for the temporary filesystem
        used to store data prior to zipping.
    :type temp_fs: str

    """

    def __new__(cls,
                file,
                write=False,
                compression=zipfile.ZIP_DEFLATED,
                encoding="utf-8",
                temp_fs="temp://__ziptemp__"):
        # This magic returns a different class instance based on the
        # value of the ``write`` parameter.
        if write:
            return WriteZipFS(file,
                              compression=compression,
                              encoding=encoding,
                              temp_fs=temp_fs)
        else:
            return ReadZipFS(file, encoding=encoding)


@six.python_2_unicode_compatible
class WriteZipFS(WrapFS):
    """A writable zip file."""

    def __init__(self,
                 file,
                 compression=zipfile.ZIP_DEFLATED,
                 encoding="utf-8",
                 temp_fs="temp://__ziptemp__"):
        self._file = file
        self.compression = compression
        self.encoding = encoding
        self._temp_fs_url = temp_fs
        self._temp_fs = open_fs(temp_fs)
        self._meta = self._temp_fs.getmeta().copy()
        super(WriteZipFS, self).__init__(self._temp_fs)

    def __repr__(self):
        t = "WriteZipFS({!r}, compression={!r}, encoding={!r}, temp_fs={!r})"
        return t.format(
            self._file,
            self.compression,
            self.encoding,
            self._temp_fs_url
        )

    def __str__(self):
        return "<zipfs-write '{}'>".format(self._file)

    def delegate_path(self, path):
        return self._temp_fs, path

    def delegate_fs(self):
        return self._temp_fs

    def close(self):
        if not self.isclosed():
            try:
                self.write_zip()
            finally:
                self._temp_fs.close()
        super(WriteZipFS, self).close()

    def write_zip(self, file=None, compression=None, encoding=None):
        """
        Write zip to a file.

        .. note ::
            This is called automatically when the ZipFS is closed.

        :param file: Destination file, may be a file name or an open
            file object.
        :type file: str or file-like
        :param compression: Compression to use (one of the constants
            defined in the zipfile module in the stdlib).

        """
        if not self.isclosed():
            write_zip(
                self._temp_fs,
                file or self._file,
                compression=compression or self.compression,
                encoding=encoding or self.encoding
            )


@six.python_2_unicode_compatible
class ReadZipFS(FS):
    """A readable zip file."""

    _meta = {
        'case_insensitive': True,
        'network': False,
        'read_only': True,
        'supports_rename': False,
        'thread_safe': True,
        'unicode_paths': True,
        'virtual': False,
    }

    def __init__(self, file, encoding='utf-8'):
        super(ReadZipFS, self).__init__()
        self._file = file
        self.encoding = encoding
        self._zip = zipfile.ZipFile(file, 'r')
        self._directory_fs = None

    def __repr__(self):
        return "ReadZipFS({!r})".format(self._file)

    def __str__(self):
        return "<zipfs '{}'>".format(self._file)

    def _path_to_zip_name(self, path):
        """Convert a path to a zip file name."""
        if self._directory.isdir(path):
            return relpath(normpath(path)) + '/'
        else:
            return relpath(normpath(path))

    @property
    def _directory(self):
        """
        Make a memory filesystem with the same directory structure
        as the zip.

        """
        self.check()
        with self._lock:
            if self._directory_fs is None:
                self._directory_fs = _fs = MemoryFS()
                for zip_name in self._zip.namelist():
                    resource_name = zip_name
                    if six.PY2:
                        resource_name =\
                            resource_name.decode(self.encoding, 'replace')
                    if resource_name.endswith('/'):
                        _fs.makedirs(resource_name, recreate=True)
                    else:
                        _fs.makedirs(dirname(resource_name), recreate=True)
                        _fs.create(resource_name)
            return self._directory_fs

    def getinfo(self, path, namespaces=None):
        self.check()
        namespaces = namespaces or ()
        _path = normpath(path)
        if _path == '/':
            raw_info = {
                "basic":
                {
                    "name": "",
                    "is_dir": True,
                },
                "details":
                {
                    "type": int(ResourceType.directory)
                }
            }
        else:
            basic_info = self._directory.getinfo(_path)
            zip_name = self._path_to_zip_name(path)
            try:
                zip_info = self._zip.getinfo(zip_name)
            except KeyError:
                # Can occur if there is an implied directory in the zip
                raw_info = {
                    "basic":
                    {
                        "name": basic_info.name,
                        "is_dir": basic_info.is_dir
                    }
                }
            else:
                modified_epoch = datetime_to_epoch(
                    datetime(*zip_info.date_time)
                )
                raw_zip_info = {
                    k: getattr(zip_info, k)
                    for k in dir(zip_info)
                    if (not k.startswith('_') and
                        not callable(getattr(zip_info, k)))
                }
                raw_info = {
                    "basic":
                    {
                        "name": basic_info.name,
                        "is_dir": basic_info.is_dir,
                    },
                    "details":
                    {
                        "size": zip_info.file_size,
                        "type": int(
                            ResourceType.directory
                            if basic_info.is_dir else
                            ResourceType.file
                        ),
                        "modified": modified_epoch
                    },
                    "zip": raw_zip_info
                }

        return Info(raw_info)

    def setinfo(self, path, info):
        self.check()
        raise errors.ResourceReadOnly(path)

    def listdir(self, path):
        self.check()
        return self._directory.listdir(path)

    def makedir(self, path, permissions=None, recreate=False):
        self.check()
        raise errors.ResourceReadOnly(path)

    def openbin(self, path, mode="r", buffering=-1, **kwargs):
        self.check()
        if 'w' in mode or '+' in mode or 'a' in mode:
            raise errors.ResourceReadOnly(path)

        if not self._directory.exists(path):
            raise errors.ResourceNotFound(path)
        elif self._directory.isdir(path):
            raise errors.FileExpected(path)

        zip_name = self._path_to_zip_name(path)
        bin_file = self._zip.open(zip_name, 'r')
        return RawWrapper(bin_file)

    def remove(self, path):
        self.check()
        raise errors.ResourceReadOnly(path)

    def removedir(self, path):
        self.check()
        raise errors.ResourceReadOnly(path)

    def close(self):
        super(ReadZipFS, self).close()
        self._zip.close()

    def getbytes(self, path):
        self.check()
        if not self._directory.isfile(path):
            raise errors.ResourceNotFound(path)
        zip_name = self._path_to_zip_name(path)
        zip_bytes = self._zip.read(zip_name)
        return zip_bytes


if __name__ == "__main__":  # pragma: nocover
    from fs.tree import render
    from fs.opener import open_fs

    with ZipFS('tests.zip') as zip_fs:
        print(zip_fs.listdir('/'))
        print(zip_fs.listdir('/tests/'))
        print(zip_fs.gettext('tests/ttt/settings.ini'))
        render(zip_fs)
        print(zip_fs)
        print(repr(zip_fs))

    with ZipFS("zipfs.zip", write=True) as zip_fs:
        zip_fs.makedirs('foo/bar')
        zip_fs.settext('foo/bar/baz.txt', 'Hello, World')
        print(zip_fs)
        print(repr(zip_fs))



# -*- coding: utf-8 -*-

"""
Render a FS object as text tree views.

"""

from __future__ import print_function
from __future__ import unicode_literals

import sys

from fs.path import abspath, join, normpath


def render(fs,
           path='/',
           file=None,
           encoding=None,
           max_levels=5,
           with_color=None,
           dirs_first=True,
           exclude=None,
           filter=None):
    """
    Render a directory structure in to a pretty tree.

    :param fs: A filesystem.
    :type fs: A :class:`~fs.base.FS` instance
    :param file: An open file-like object to render the tree, or
        ``None`` for stdout.
    :type file: file or None
    :type encoding: Unicode encoding, or None to auto-detect.
    :type encoding: str or None
    :param int max_levels: Maximum number of levels to display, or None
        for no maximum.
    :param bool with_color: Enable terminal color output, or None to
        auto-detect terminal.
    :param bool dirs_first: Show directories first.
    :param list exclude: Option list of directory patterns to exclude
        from the tree render.
    :param filter: Optional list of files patterns to match in the tree
        render.
    :rtype: tuple
    :returns: A tuple of ``(<directory count>, <file count>)``.

    """
    file = file or sys.stdout
    if encoding is None:
        encoding = getattr(file, 'encoding', 'utf-8') or 'utf-8'
    is_tty = hasattr(file, 'isatty') and file.isatty()

    if with_color is None:
        is_windows = sys.platform.startswith('win')
        with_color = False if is_windows else is_tty

    if encoding.lower() == 'utf-8' and with_color:
        char_vertline = '│'
        char_newnode = '├'
        char_line = '──'
        char_corner = '└'
    else:
        char_vertline = '|'
        char_newnode = '|'
        char_line = '--'
        char_corner = '`'

    indent = ' ' * 4
    line_indent = char_vertline + ' ' * 3

    def write(line):
        """Write a line to the output."""
        print(line, file=file)

    def format_prefix(prefix):
        """Format the prefix lines."""
        if not with_color:
            return prefix
        return '\x1b[32m%s\x1b[0m' % prefix

    def format_dirname(dirname):
        """Format a directory name."""
        if not with_color:
            return dirname
        return '\x1b[1;34m%s\x1b[0m' % dirname

    def format_error(msg):
        """Format an error."""
        if not with_color:
            return msg
        return '\x1b[31m%s\x1b[0m' % msg

    def format_filename(fname):
        """Format a filename."""
        if not with_color:
            return fname
        if fname.startswith('.'):
            fname = '\x1b[33m%s\x1b[0m' % fname
        return fname

    def sort_key_dirs_first(info):
        """Sort key func with directories first."""
        return (not info.is_dir, info.name.lower())

    def sort_key(info):
        """Default  key for info."""
        return info.name.lower()

    counts = {"dirs": 0, "files": 0}

    def format_directory(path, levels):
        """Recursive directory function."""
        try:
            directory = sorted(
                fs.filterdir(path, exclude_dirs=exclude, files=filter),
                key=sort_key_dirs_first if dirs_first else sort_key
            )
        except Exception as error:
            prefix = ''.join(
                indent if last else line_indent
                for last in levels
            ) + char_corner + char_line
            write("{} {}".format(
                format_prefix(prefix),
                format_error("error ({})".format(error))
            ))
            return
        _last = len(directory) - 1
        for i, info in enumerate(directory):
            is_last_entry = i == _last
            counts['dirs' if info.is_dir else 'files'] += 1
            prefix = ''.join(
                indent if last else line_indent
                for last in levels
            )
            prefix += char_corner if is_last_entry else char_newnode
            if info.is_dir:
                write("{} {}".format(
                    format_prefix(prefix + char_line),
                    format_dirname(info.name)
                ))
                if max_levels is None or len(levels) < max_levels:
                    format_directory(
                        join(path, info.name),
                        levels + [is_last_entry]
                    )
            else:
                write("{} {}".format(
                    format_prefix(prefix + char_line),
                    format_filename(info.name)
                ))

    format_directory(abspath(normpath(path)), [])
    return counts['dirs'], counts['files']



"""
This module can compress the contents of a filesystem.

Currently zip and tar are supported.
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from datetime import datetime
import time
import zipfile
import tarfile

import six

from .enums import ResourceType
from .path import relpath
from .time import datetime_to_epoch
from .errors import NoSysPath
from .walk import Walker


def write_zip(src_fs,
              file,
              compression=zipfile.ZIP_DEFLATED,
              encoding="utf-8",
              walker=None):
    """
    Write the contents of a filesystem to a zip file.

    :param file: Destination file, may be a file name or an open file
        object.
    :type file: str or file-like.
    :param compression: Compression to use (one of the constants defined
        in the zipfile module in the stdlib).
    :type compression: str
    :param encoding: The encoding to use for filenames. The default is
        ``"utf-8"``, use ``"CP437"`` if compatibility with WinZip is
        desired.
    :type encoding: str
    :param walker: A :class:`~fs.walk.Walker` instance, or None to use
        default walker. You can use this to specify which files you
        want to compress.
    :type walker: Walker or None

    """
    _zip = zipfile.ZipFile(
        file,
        mode="w",
        compression=compression,
        allowZip64=True
    )
    walker = walker or Walker()
    with _zip:
        gen_walk = walker.info(src_fs, namespaces=["details", "stat"])
        for path, info in gen_walk:
            # Zip names must be relative, directory names must end
            # with a slash.
            zip_name = relpath(path + '/' if info.is_dir else path)
            if not six.PY3:
                # Python2 expects bytes filenames
                zip_name = zip_name.encode(encoding, 'replace')

            if info.has_namespace('stat'):
                # If the file has a stat namespace, get the
                # zip time directory from the stat structure
                st_mtime = info.get('stat', 'st_mtime', None)
                _mtime = time.localtime(st_mtime)
                zip_time = _mtime[0:6]
            else:
                # Otherwise, use the modified time from details
                # namespace.
                mt = info.modified or datetime.utcnow()
                zip_time = (
                    mt.year, mt.month, mt.day,
                    mt.hour, mt.minute, mt.second
                )
            zip_info = zipfile.ZipInfo(zip_name, zip_time)

            if info.is_dir:
                # This is how to record directories with zipfile
                _zip.writestr(zip_info, b'')
            else:
                # Get a syspath if possible
                try:
                    sys_path = src_fs.getsyspath(path)
                except NoSysPath:
                    # Write from bytes
                    _zip.writestr(
                        zip_info,
                        src_fs.getbytes(path)
                    )
                else:
                    # Write from a file which is (presumably)
                    # more memory efficient
                    _zip.write(sys_path, zip_name)



def write_tar(src_fs,
              file,
              compression=None,
              encoding="utf-8",
              walker=None):
    """
    Write the contents of a filesystem to a tar file.

    :param file: Destination file, may be a file name or an open file
        object.
    :type file: str or file-like.
    :param compression: Compression to use.
    :type compression: str
    :param encoding: The encoding to use for filenames. The default is
        ``"utf-8"``.
    :type encoding: str
    :param walker: A :class:`~fs.walk.Walker` instance, or None to use
        default walker. You can use this to specify which files you
        want to compress.
    :type walker: Walker or None

    """

    type_map = {
        ResourceType.block_special_file: tarfile.BLKTYPE,
        ResourceType.character: tarfile.CHRTYPE,
        ResourceType.directory: tarfile.DIRTYPE,
        ResourceType.fifo: tarfile.FIFOTYPE,
        ResourceType.file: tarfile.REGTYPE,
        ResourceType.socket: tarfile.AREGTYPE,   # no type for socket
        ResourceType.symlink: tarfile.SYMTYPE,
        ResourceType.unknown: tarfile.AREGTYPE,  # no type for unknown
    }

    tar_attr = [
        ('uid', 'uid'),
        ('gid', 'gid'),
        ('uname', 'user'),
        ('gname', 'group'),
    ]

    mode = 'w:{}'.format(compression or '')
    if hasattr(file, 'write'):
        _tar = tarfile.open(fileobj=file, mode=mode)
    else:
        _tar = tarfile.open(file, mode=mode)

    current_time = time.time()
    walker = walker or Walker()
    with _tar:
        gen_walk = walker.info(src_fs, namespaces=["details", "stat", "access"])
        for path, info in gen_walk:
            # Tar names must be relative
            tar_name = relpath(path)
            if not six.PY3:
                # Python2 expects bytes filenames
                tar_name = tar_name.encode(encoding, 'replace')

            tar_info = tarfile.TarInfo(tar_name)

            if info.has_namespace('stat'):
                mtime = info.get('stat', 'st_mtime', current_time)
            else:
                mtime = info.modified or current_time

            if isinstance(mtime, datetime):
                mtime = datetime_to_epoch(mtime)
            if isinstance(mtime, float):
                mtime = int(mtime)
            tar_info.mtime = mtime

            for tarattr, infoattr in tar_attr:
                if getattr(info, infoattr, None) is not None:
                    setattr(tar_info, tarattr, getattr(info, infoattr, None))

            if info.has_namespace('access'):
                tar_info.mode = getattr(info.permissions, 'mode', 0o420)

            if info.is_dir:
                tar_info.type = tarfile.DIRTYPE
                _tar.addfile(tar_info)
            else:
                tar_info.type = type_map.get(info.type, tarfile.REGTYPE)
                tar_info.size = info.size
                with src_fs.openbin(path) as bin_file:
                    _tar.addfile(tar_info, bin_file)



"""

Defines the Exception classes thrown by PyFilesystem objects.

Errors relating to the underlying filesystem are translated in to one of
the following exceptions.

All Exception classes are derived from :class:`~fs.errors.FSError` which
may be used as a catch-all filesystem exception.

"""

from __future__ import unicode_literals
from __future__ import print_function

import six
from six import text_type

__all__ = [
    'CreateFailed',
    'DestinationExists',
    'DirectoryExists',
    'DirectoryExpected',
    'DirectoryNotEmpty',
    'FileExists',
    'FileExpected',
    'FilesystemClosed',
    'FSError',
    'IllegalBackReference',
    'InsufficientStorage',
    'InvalidCharsInPath',
    'InvalidPath',
    'MissingInfoNamespace',
    'NoSysPath',
    'NoURL',
    'OperationFailed',
    'OperationTimeout',
    'PathError',
    'PermissionDenied',
    'RemoteConnectionError',
    'RemoveRootError',
    'ResourceError',
    'ResourceInvalid',
    'ResourceLocked',
    'ResourceNotFound',
    'ResourceReadOnly',
    'Unsupported',
]


class MissingInfoNamespace(AttributeError):
    """Raised when an expected namespace was missing."""

    def __init__(self, namespace):
        msg = "namespace '{}' is required for this attribute"
        super(MissingInfoNamespace, self).__init__(
            msg.format(namespace)
        )


@six.python_2_unicode_compatible
class FSError(Exception):
    """Base exception class for the FS module."""

    default_message = "Unspecified error"

    def __init__(self, msg=None):
        self._msg = msg or self.default_message
        super(FSError, self).__init__()

    def __str__(self):
        """The error message."""
        msg = self._msg.format(**self.__dict__)
        return msg

    def __repr__(self):
        msg = self._msg.format(**self.__dict__)
        return "{}({!r})".format(self.__class__.__name__, msg)


class FilesystemClosed(FSError):
    """An exception thrown when attempting to use a closed filesystem."""

    default_message = "attempt to use closed filesystem"


class CreateFailed(FSError):
    """An exception thrown when a FS could not be created."""

    default_message = "unable to create filesystem"


class PathError(FSError):
    """Exception for errors to do with a path string."""

    default_message = "path '{path}' is invalid"

    def __init__(self, path, msg=None):
        self.path = path
        super(PathError, self).__init__(msg=msg)


class NoSysPath(PathError):
    """Exception raised when there is no sys path."""

    default_message = "path '{path}' does not map to the local filesystem"


class NoURL(PathError):
    """Raised when there is no URL for a given path."""

    default_message = "path '{path}' has no '{purpose}' URL"

    def __init__(self, path, purpose, msg=None):
        self.purpose = purpose
        super(NoURL, self).__init__(path, msg=msg)


class InvalidPath(PathError):
    """Base exception for fs paths that can't be mapped on to the
    underlaying filesystem."""

    default_message = "path '{path}' is invalid on this filesystem "


class InvalidCharsInPath(InvalidPath):
    """The path contains characters that are invalid on this filesystem."""

    default_message = "path '{path}' contains invalid characters"


class OperationFailed(FSError):
    """Base exception class for errors associated with a specific operation."""

    default_message = "operation failed, {details}"

    def __init__(self, path=None, exc=None, msg=None):
        self.path = path
        self.exc = exc
        self.details = '' if exc is None else text_type(exc)
        self.errno = getattr(exc, "errno", None)
        super(OperationFailed, self).__init__(msg=msg)


class Unsupported(OperationFailed):
    """Exception raised for operations that are not supported by the FS."""

    default_message = "not supported"


class RemoteConnectionError(OperationFailed):
    """Exception raised when operations encounter remote connection trouble."""

    default_message = "remote connection error"


class InsufficientStorage(OperationFailed):
    """Exception raised when operations encounter storage space trouble."""

    default_message = "insufficient storage space"


class PermissionDenied(OperationFailed):
    """Permissions error."""

    default_message = "permission denied"


class OperationTimeout(OperationFailed):
    """Filesystem took too long."""

    default_message = "operation timed out"


class RemoveRootError(OperationFailed):
    """Attempt to remove the root directory."""

    default_message = "root directory may not be removed"


class ResourceError(FSError):
    """Base exception class for error associated with a specific resource."""

    default_message = "failed on path {path}"

    def __init__(self, path, exc=None, msg=None):
        self.path = path
        self.exc = exc
        super(ResourceError, self).__init__(msg=msg)


class ResourceNotFound(ResourceError):
    """Exception raised when a required resource is not found."""

    default_message = "resource '{path}' not found"


class ResourceInvalid(ResourceError):
    """Exception raised when a resource is the wrong type."""

    default_message = "resource '{path}' is invalid for this operation"


class FileExists(ResourceError):
    """Exception raises when opening a file in exclusive mode."""

    default_message = "resource '{path}' exists"


class FileExpected(ResourceInvalid):
    """Exception raises when a file was expected."""

    default_message = "path '{path}' should be a file"


class DirectoryExpected(ResourceInvalid):
    """Exception raises when a directory was expected."""

    default_message = "path '{path}' should be a directory"


class DestinationExists(ResourceError):
    """Exception raised when a target destination already exists."""

    default_message = "destination '{path}' exists"


class DirectoryExists(ResourceError):
    """Exception raised when trying to make a directory that already
    exists."""

    default_message = "directory '{path}' exists"


class DirectoryNotEmpty(ResourceError):
    """Exception raised when a directory to be removed is not empty."""

    default_message = "directory '{path}' is not empty"


class ResourceLocked(ResourceError):
    """Exception raised when a resource can't be used because it is locked."""

    default_message = "resource '{path}' is locked"


class ResourceReadOnly(ResourceError):
    """Raised when attempting to modify a read only resource."""

    default_message = "resource '{path}' is read only"


class IllegalBackReference(ValueError):
    """
    Exception raised when too many backrefs exist in a path.

    This error will occur if the back references in a path would be
    outside of the root. For example, ``"/foo/../../"``, contains two back
    references which would reference a directory above the root.

    .. note::

        This exception is a subclass of ``ValueError`` as it is not
        strictly speaking an issue with a filesystem or resource.

    """

    def __init__(self, path):
        _msg = \
            "path '{path}' contains back-references outside of filesystem"
        _msg = _msg.format(path=path)
        super(IllegalBackReference, self).__init__(_msg)



__import__('pkg_resources').declare_namespace(__name__)

from ._version import __version__
from .enums import ResourceType, Seek
from .opener import open_fs

__all__ = ['__version__', 'ResourceType', 'Seek', 'open_fs']



#!/usr/bin/env python

from setuptools import setup, find_packages

with open('fs/_version.py') as f:
    exec(f.read())

CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Topic :: System :: Filesystems',
]

with open('README.rst', 'rt') as f:
    DESCRIPTION = f.read()

REQUIREMENTS = [
    "appdirs~=1.4.3",
    "pytz",
    "setuptools",
    "six~=1.10",
]

setup(
    author="Will McGugan",
    author_email="will@willmcgugan.com",
    classifiers=CLASSIFIERS,
    description="Python's filesystem abstraction layer",
    install_requires=REQUIREMENTS,
    extras_require={
        "scandir :python_version < '3.5'": ['scandir~=1.5'],
        ":python_version < '3.4'": ['enum34~=1.1.6']
    },
    entry_points={'fs.opener': [
        'ftp  = fs.opener.ftpfs:FTPOpener',
        'file = fs.opener.osfs:OSFSOpener',
        'osfs = fs.opener.osfs:OSFSOpener',
        'mem  = fs.opener.memoryfs:MemOpener',
        'tar  = fs.opener.tarfs:TarOpener',
        'temp = fs.opener.tempfs:TempOpener',
        'zip  = fs.opener.zipfs:ZipOpener',

        'userdata = fs.opener.appfs:AppFSOpener',
        'userconf = fs.opener.appfs:AppFSOpener',
        'sitedata = fs.opener.appfs:AppFSOpener',
        'siteconf = fs.opener.appfs:AppFSOpener',
        'usercache = fs.opener.appfs:AppFSOpener',
        'userlog = fs.opener.appfs:AppFSOpener',
    ]},
    license="MIT",
    long_description=DESCRIPTION,
    name='fs',
    packages=find_packages(exclude=("tests",)),
    platforms=['any'],
    test_suite="nose.collector",
    tests_require=['appdirs', 'mock', 'pytz', 'pyftpdlib'],
    url="https://github.com/PyFilesystem/pyfilesystem2",
    version=__version__,
)



from __future__ import unicode_literals

import unittest

from fs import errors
from fs import open_fs
from fs import wrap


class TestWrap(unittest.TestCase):

    def test_readonly(self):
        mem_fs = open_fs('mem://')
        fs = wrap.read_only(mem_fs)

        with self.assertRaises(errors.ResourceReadOnly):
            fs.open('foo', 'w')

        with self.assertRaises(errors.ResourceReadOnly):
            fs.appendtext('foo', 'bar')

        with self.assertRaises(errors.ResourceReadOnly):
            fs.appendbytes('foo', b'bar')

        with self.assertRaises(errors.ResourceReadOnly):
            fs.makedir('foo')

        with self.assertRaises(errors.ResourceReadOnly):
            fs.move('foo', 'bar')

        with self.assertRaises(errors.ResourceReadOnly):
            fs.openbin('foo', 'w')

        with self.assertRaises(errors.ResourceReadOnly):
            fs.remove('foo')

        with self.assertRaises(errors.ResourceReadOnly):
            fs.removedir('foo')

        with self.assertRaises(errors.ResourceReadOnly):
            fs.setinfo('foo', {})

        with self.assertRaises(errors.ResourceReadOnly):
            fs.settimes('foo', {})

        with self.assertRaises(errors.ResourceReadOnly):
            fs.copy('foo', 'bar')

        with self.assertRaises(errors.ResourceReadOnly):
            fs.create('foo')

        with self.assertRaises(errors.ResourceReadOnly):
            fs.settext('foo', 'bar')

        with self.assertRaises(errors.ResourceReadOnly):
            fs.setbytes('foo', b'bar')

        with self.assertRaises(errors.ResourceReadOnly):
            fs.makedirs('foo/bar')

        with self.assertRaises(errors.ResourceReadOnly):
            fs.touch('foo')

        with self.assertRaises(errors.ResourceReadOnly):
            fs.setbinfile('foo', None)

        with self.assertRaises(errors.ResourceReadOnly):
            fs.setfile('foo', None)

        self.assertTrue(mem_fs.isempty('/'))
        mem_fs.setbytes('file', b'read me')
        with fs.openbin('file') as read_file:
            self.assertEqual(read_file.read(), b'read me')

        with fs.open('file', 'rb') as read_file:
            self.assertEqual(read_file.read(), b'read me')

    def test_cachedir(self):
        mem_fs = open_fs('mem://')
        mem_fs.makedirs('foo/bar/baz')
        mem_fs.touch('egg')

        fs = wrap.cache_directory(mem_fs)
        self.assertEqual(
            sorted(fs.listdir('/')),
            ['egg', 'foo']
        )
        self.assertEqual(
            sorted(fs.listdir('/')),
            ['egg', 'foo']
        )
        self.assertTrue(fs.isdir('foo'))
        self.assertTrue(fs.isdir('foo'))
        self.assertTrue(fs.isfile('egg'))
        self.assertTrue(fs.isfile('egg'))

        self.assertEqual(fs.getinfo('foo'), mem_fs.getinfo('foo'))
        self.assertEqual(fs.getinfo('foo'), mem_fs.getinfo('foo'))

        self.assertEqual(fs.getinfo('/'), mem_fs.getinfo('/'))
        self.assertEqual(fs.getinfo('/'), mem_fs.getinfo('/'))

        with self.assertRaises(errors.ResourceNotFound):
            fs.getinfo('/foofoo')




from __future__ import unicode_literals

import unittest

from six import text_type

from fs.mode import check_readable, check_writable, Mode


class TestMode(unittest.TestCase):

    def test_checks(self):
        self.assertTrue(check_readable('r'))
        self.assertTrue(check_readable('r+'))
        self.assertTrue(check_readable('rt'))
        self.assertTrue(check_readable('rb'))

        self.assertFalse(check_readable('w'))
        self.assertTrue(check_readable('w+'))
        self.assertFalse(check_readable('wt'))
        self.assertFalse(check_readable('wb'))
        self.assertFalse(check_readable('a'))

        self.assertTrue(check_writable('w'))
        self.assertTrue(check_writable('w+'))
        self.assertTrue(check_writable('r+'))
        self.assertFalse(check_writable('r'))
        self.assertTrue(check_writable('a'))

    def test_mode_object(self):
        with self.assertRaises(ValueError):
            Mode('')
        with self.assertRaises(ValueError):
            Mode('J')
        with self.assertRaises(ValueError):
            Mode('b')
        with self.assertRaises(ValueError):
            Mode('rtb')

        mode = Mode('w')
        repr(mode)
        self.assertEqual(text_type(mode), 'w')
        self.assertTrue(mode.create)
        self.assertFalse(mode.reading)
        self.assertTrue(mode.writing)
        self.assertFalse(mode.appending)
        self.assertFalse(mode.updating)
        self.assertTrue(mode.truncate)
        self.assertFalse(mode.exclusive)
        self.assertFalse(mode.binary)
        self.assertTrue(mode.text)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import errno
import io
import itertools
import logging
import os
import platform
import stat
import sys

import six

try:
    from os import scandir
except ImportError:
    try:
        from scandir import scandir
    except ImportError:  # pragma: nocover
        scandir = None

from . import errors
from .errors import FileExists
from .base import FS
from .enums import ResourceType
from ._fscompat import fsencode, fsdecode, fspath
from .info import Info
from .path import basename
from .permissions import Permissions
from .error_tools import convert_os_errors
from .mode import Mode, validate_open_mode
from .errors import NoURL

log = logging.getLogger('fs.osfs')


_WINDOWS_PLATFORM = platform.system() == 'Windows'


@six.python_2_unicode_compatible
class OSFS(FS):
    """
    Create an OSFS.

    :param root_path: An OS path or path-like object to the location on
        your HD you wish to manage.
    :type root_path: str or path-like
    :param create: Set to ``True`` to create the root directory if it
        does not already exist, otherwise the directory should exist
        prior to creating the ``OSFS`` instance.
    :type create: bool
    :param int create_mode: The permissions that will be used to create
        the directory if ``create`` is True and the path doesn't exist,
        defaults to ``0o777``.
    :raises `fs.errors.CreateFailed`: If ``root_path`` does not
        exists, or could not be created.


    Here are some examples of creating ``OSFS`` objects::

        current_directory_fs = OSFS('.')
        home_fs = OSFS('~/')
        windows_system32_fs = OSFS('c://system32')

    """

    def __init__(self,
                 root_path,
                 create=False,
                 create_mode=0o777):
        """Create an OSFS instance."""

        super(OSFS, self).__init__()
        root_path = fsdecode(fspath(root_path))
        _root_path = os.path.expanduser(os.path.expandvars(root_path))
        _root_path = os.path.normpath(os.path.abspath(_root_path))
        self.root_path = _root_path

        if create:
            try:
                if not os.path.isdir(_root_path):
                    os.makedirs(_root_path, mode=create_mode)
            except OSError as error:
                raise errors.CreateFailed(
                    'unable to create {} ({})'.format(root_path, error)
                )
        else:
            if not os.path.isdir(_root_path):
                raise errors.CreateFailed(
                    'root path does not exists'
                )

        _meta = self._meta = {
            'case_insensitive': os.path.normcase('Aa') != 'aa',
            'network': False,
            'read_only': False,
            'supports_rename': True,
            'thread_safe': True,
            'unicode_paths': os.path.supports_unicode_filenames,
            'virtual': False,
        }

        if _WINDOWS_PLATFORM:  # pragma: nocover
            _meta["invalid_path_chars"] =\
                ''.join(six.unichr(n) for n in range(31)) + '\\:*?"<>|'
        else:
            _meta["invalid_path_chars"] = '\0'

            if 'PC_PATH_MAX' in os.pathconf_names:
                _meta['max_sys_path_length'] = (
                    os.pathconf(
                        fsencode(_root_path),
                        os.pathconf_names['PC_PATH_MAX']
                    )
                )

    def __repr__(self):
        _fmt = "{}({!r})"
        return _fmt.format(self.__class__.__name__,
                           self.root_path)

    def __str__(self):
        fmt = "<{} '{}'>"
        return fmt.format(self.__class__.__name__.lower(),
                          self.root_path)

    def _to_sys_path(self, path):
        """Convert a FS path to a path on the OS."""
        sys_path = os.path.join(
            self.root_path,
            path.lstrip('/').replace('/', os.sep)
        )
        return sys_path

    @classmethod
    def _make_details_from_stat(cls, stat_result):
        """Make an info dict from a stat_result object."""
        details = {
            '_write': ['accessed', 'modified'],
            'accessed': stat_result.st_atime,
            'modified': stat_result.st_mtime,
            'size': stat_result.st_size,
            'type': int(cls._get_type_from_stat(stat_result))
        }
        # On other Unix systems (such as FreeBSD), the following
        # attributes may be available (but may be only filled out if
        # root tries to use them):
        details['created'] = getattr(stat_result, 'st_birthtime', None)
        ctime_key = (
            'created'
            if _WINDOWS_PLATFORM
            else 'metadata_changed'
        )
        details[ctime_key] = stat_result.st_ctime
        return details

    @classmethod
    def _make_access_from_stat(cls, stat_result):
        access = {}
        access['permissions'] = Permissions(
            mode=stat_result.st_mode
        ).dump()
        access['gid'] = stat_result.st_gid
        access['uid'] = stat_result.st_uid
        if not _WINDOWS_PLATFORM:
            import grp
            import pwd
            try:
                access['group'] = grp.getgrgid(access['gid']).gr_name
            except KeyError:  # pragma: nocover
                pass

            try:
                access['user'] = pwd.getpwuid(access['uid']).pw_name
            except KeyError:  # pragma: nocover
                pass
        return access

    STAT_TO_RESOURCE_TYPE = {
        stat.S_IFDIR: ResourceType.directory,
        stat.S_IFCHR: ResourceType.character,
        stat.S_IFBLK: ResourceType.block_special_file,
        stat.S_IFREG: ResourceType.file,
        stat.S_IFIFO: ResourceType.fifo,
        stat.S_IFLNK: ResourceType.symlink,
        stat.S_IFSOCK: ResourceType.socket
    }

    @classmethod
    def _get_type_from_stat(cls, _stat):
        """Get the resource type from a stat_result object."""
        st_mode = _stat.st_mode
        st_type = stat.S_IFMT(st_mode)
        return cls.STAT_TO_RESOURCE_TYPE.get(st_type, ResourceType.unknown)

    # --------------------------------------------------------
    # Required Methods
    # --------------------------------------------------------

    def _gettarget(self, sys_path):
        try:
            target = os.readlink(sys_path)
        except OSError:
            return None
        else:
            return target

    def _make_link_info(self, sys_path):
        _target = self._gettarget(sys_path)
        link = {
            'target': _target,
        }
        return link

    def getinfo(self, path, namespaces=None):
        self.check()
        namespaces = namespaces or ()
        _path = self.validatepath(path)
        sys_path = self.getsyspath(_path)
        _lstat = None
        with convert_os_errors('getinfo', path):
            _stat = os.stat(sys_path)
            if 'lstat' in namespaces:
                _stat = os.lstat(sys_path)

        info = {
            'basic': {
                'name': basename(_path),
                'is_dir': stat.S_ISDIR(_stat.st_mode)
            }
        }
        if 'details' in namespaces:
            info['details'] = self._make_details_from_stat(_stat)
        if 'stat' in namespaces:
            info['stat'] = {
                k: getattr(_stat, k)
                for k in dir(_stat) if k.startswith('st_')
            }
        if 'lstat' in namespaces:
            info['lstat'] = {
                k: getattr(_lstat, k)
                for k in dir(_lstat) if k.startswith('st_')
            }
        if 'link' in namespaces:
            info['link'] = self._make_link_info(sys_path)
        if 'access' in namespaces:
            info['access'] = self._make_access_from_stat(_stat)

        return Info(info)

    def listdir(self, path):
        self.check()
        _path = self.validatepath(path)
        sys_path = self._to_sys_path(_path)
        with convert_os_errors('listdir', path, directory=True):
            names = os.listdir(sys_path)
        return names

    def makedir(self, path, permissions=None, recreate=False):
        self.check()
        mode = Permissions.get_mode(permissions)
        _path = self.validatepath(path)
        sys_path = self._to_sys_path(_path)
        with convert_os_errors('makedir', path, directory=True):
            try:
                os.mkdir(sys_path, mode)
            except OSError as error:
                if error.errno == errno.ENOENT:
                    raise errors.ResourceNotFound(path)
                elif error.errno == errno.EEXIST and recreate:
                    pass
                else:
                    raise
            return self.opendir(_path)

    def openbin(self, path, mode="r", buffering=-1, **options):
        _mode = Mode(mode)
        _mode.validate_bin()
        self.check()
        _path = self.validatepath(path)
        sys_path = self._to_sys_path(_path)
        with convert_os_errors('openbin', path):
            if six.PY2 and _mode.exclusive and self.exists(path):
                raise errors.FileExists(path)
            binary_file = io.open(
                sys_path,
                mode=_mode.to_platform_bin(),
                buffering=buffering,
                **options
            )
        return binary_file

    def remove(self, path):
        self.check()
        _path = self.validatepath(path)
        sys_path = self._to_sys_path(_path)
        with convert_os_errors('remove', path):
            try:
                os.remove(sys_path)
            except OSError as error:
                if error.errno == errno.EACCES and sys.platform == "win32":
                    # sometimes windows says this for attempts to remove a dir
                    if os.path.isdir(sys_path):  # pragma: nocover
                        raise errors.FileExpected(path)
                if error.errno == errno.EPERM and sys.platform == "darwin":
                    # sometimes OSX says this for attempts to remove a dir
                    if os.path.isdir(sys_path):  # pragma: nocover
                        raise errors.FileExpected(path)
                raise

    def removedir(self, path):
        self.check()
        _path = self.validatepath(path)
        if _path == '/':
            raise errors.RemoveRootError()
        sys_path = self._to_sys_path(path)
        with convert_os_errors('removedir', path, directory=True):
            os.rmdir(sys_path)

    # --------------------------------------------------------
    # Optional Methods
    # --------------------------------------------------------

    def getsyspath(self, path):
        sys_path = self._to_sys_path(path)
        return sys_path

    def geturl(self, path, purpose='download'):
        if purpose != 'download':
            raise NoURL(path, purpose)
        return "file://" + self.getsyspath(path)

    def gettype(self, path):
        self.check()
        sys_path = self._to_sys_path(path)
        with convert_os_errors('gettype', path):
            stat = os.stat(sys_path)
        resource_type = self._get_type_from_stat(stat)
        return resource_type

    def islink(self, path):
        self.check()
        _path = self.validatepath(path)
        sys_path = self._to_sys_path(_path)
        if not self.exists(path):
            raise errors.ResourceNotFound(path)
        with convert_os_errors('islink', path):
            return os.path.islink(sys_path)

    def open(self,
             path,
             mode="r",
             buffering=-1,
             encoding=None,
             errors=None,
             newline='',
             line_buffering=False,
             **options):
        _mode = Mode(mode)
        validate_open_mode(mode)
        self.check()
        _path = self.validatepath(path)
        sys_path = self._to_sys_path(_path)
        with convert_os_errors('open', path):
            if six.PY2 and _mode.exclusive and self.exists(path):
                raise FileExists(path)
            _encoding = encoding or 'utf-8'
            return io.open(
                sys_path,
                mode=_mode.to_platform(),
                buffering=buffering,
                encoding=None if _mode.binary else _encoding,
                errors=errors,
                newline=None if _mode.binary else newline,
                **options
            )

    def setinfo(self, path, info):
        self.check()
        _path = self.validatepath(path)
        sys_path = self._to_sys_path(_path)
        if not os.path.exists(sys_path):
            raise errors.ResourceNotFound(path)
        if 'details' in info:
            details = info['details']
            if 'accessed' in details or 'modified' in details:
                accessed = details.get("accessed")
                modified = details.get("modified", accessed)
                accessed = int(modified if accessed is None else accessed)
                modified = int(modified)
                if accessed is not None or modified is not None:
                    with convert_os_errors('setinfo', path):
                        os.utime(sys_path, (accessed, modified))

    if scandir:
        def _scandir(self, path, namespaces=None):
            self.check()
            namespaces = namespaces or ()
            _path = self.validatepath(path)
            sys_path = self._to_sys_path(_path)
            with convert_os_errors('scandir', path, directory=True):
                for dir_entry in scandir(sys_path):
                    info = {
                        "basic": {
                            "name": dir_entry.name,
                            "is_dir": dir_entry.is_dir()
                        }
                    }
                    if 'details' in namespaces:
                        stat_result = dir_entry.stat()
                        info['details'] = \
                            self._make_details_from_stat(stat_result)
                    if 'stat' in namespaces:
                        stat_result = dir_entry.stat()
                        info['stat'] = {
                            k: getattr(stat_result, k)
                            for k in dir(stat_result) if k.startswith('st_')
                        }
                    if 'lstat' in namespaces:
                        lstat_result = dir_entry.stat(follow_symlinks=False)
                        info['lstat'] = {
                            k: getattr(lstat_result, k)
                            for k in dir(lstat_result) if k.startswith('st_')
                        }
                    if 'link' in namespaces:
                        info['link'] = self._make_link_info(
                            os.path.join(sys_path, dir_entry.name)
                        )
                    if 'access' in namespaces:
                        stat_result = dir_entry.stat()
                        info['access'] =\
                            self._make_access_from_stat(stat_result)

                    yield Info(info)

    else:

        def _scandir(self, path, namespaces=None):
            self.check()
            namespaces = namespaces or ()
            _path = self.validatepath(path)
            sys_path = self._to_sys_path(_path)
            with convert_os_errors('scandir', path, directory=True):
                for entry_name in os.listdir(sys_path):
                    entry_path = os.path.join(sys_path, entry_name)
                    stat_result = os.stat(entry_path)
                    info = {
                        "basic": {
                            "name": entry_name,
                            "is_dir": stat.S_ISDIR(stat_result.st_mode),
                        }
                    }
                    if 'details' in namespaces:
                        info['details'] = \
                            self._make_details_from_stat(stat_result)
                    if 'stat' in namespaces:
                        info['stat'] = {
                            k: getattr(stat_result, k)
                            for k in dir(stat_result) if k.startswith('st_')
                        }
                    if 'lstat' in namespaces:
                        lstat_result = os.lstat(entry_path)
                        info['lstat'] = {
                            k: getattr(lstat_result, k)
                            for k in dir(lstat_result) if k.startswith('st_')
                        }
                    if 'link' in namespaces:
                        info['link'] = self._make_link_info(
                            os.path.join(sys_path, entry_name)
                        )
                    if 'access' in namespaces:
                        info['access'] =\
                            self._make_access_from_stat(stat_result)

                    yield Info(info)


    def scandir(self, path, namespaces=None, page=None):
        iter_info = self._scandir(path, namespaces=namespaces)
        if page is not None:
            start, end = page
            iter_info = itertools.islice(iter_info, start, end)
        return iter_info



"""

A SubFS represents a directory in a 'parent' filesystem.

"""

from __future__ import print_function
from __future__ import unicode_literals

import six

from .wrapfs import WrapFS
from .path import abspath, join, normpath, relpath


@six.python_2_unicode_compatible
class SubFS(WrapFS):
    """
    A sub-directory on another filesystem.

    A SubFS is a filesystem object that maps to a sub-directory of
    another filesystem. This is the object that is returned by
    :meth:`~fs.base.FS.opendir`.

    """

    def __init__(self, parent_fs, path):
        super(SubFS, self).__init__(parent_fs)
        self._sub_dir = abspath(normpath(path))

    def __repr__(self):
        return "{}({!r}, {!r})".format(
            self.__class__.__name__,
            self._wrap_fs,
            self._sub_dir
        )

    def __str__(self):
        return "{parent}{dir}".format(
            parent=self._wrap_fs,
            dir=self._sub_dir
        )

    def delegate_fs(self):
        return self._wrap_fs

    def delegate_path(self, path):
        _path = join(self._sub_dir, relpath(normpath(path)))
        return self._wrap_fs, _path


class ClosingSubFS(SubFS):
    """
    A version of SubFS which will close its parent automatically.

    """

    def close(self):
        self.delegate_fs().close()
        super(ClosingSubFS, self).close()



from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

from collections import namedtuple
from operator import itemgetter

from six import text_type

from .base import FS
from .mode import check_writable
from . import errors
from .opener import open_fs
from .path import abspath, normpath


_PrioritizedFS = namedtuple(
    '_PrioritizedFS',
    [
        'priority',
        'fs'
    ]
)


class MultiFS(FS):
    """
    A filesystem that delegates to a sequence of other filesystems.

    Operations on the MultiFS will try each 'child' filesystem in order,
    until it succeeds. In effect, creating a filesystem that combines
    the files and dirs of its children.

    """

    _meta = {
        "virtual": True,
        "read_only": False,
        "case_insensitive": False
    }

    def __init__(self, auto_close=True):
        super(MultiFS, self).__init__()

        self._auto_close = auto_close
        self.write_fs = None

        self._write_fs_name = None
        self._sort_index = 0
        self._filesystems = {}
        self._fs_sequence = None
        self._closed = False

    def __repr__(self):
        if self._auto_close:
            return "MultiFS()"
        else:
            return "MultiFS(auto_close=False)"

    def __str__(self):
        return "<multifs>"

    def add_fs(self, name, fs, write=False, priority=0):
        """
        Adds a filesystem to the MultiFS.

        :param name: A unique name to refer to the filesystem being
            added.
        :type name: str
        :param fs: The filesystem to add.
        :type fs: Filesystem or FS URL.
        :param write: If this value is True, then the ``fs`` will be
            used as the writeable FS.
        :type write: bool
        :param priority: An integer that denotes the priority of the
            filesystem being added. Filesystems will be searched in
            descending priority order and then by the reverse order they
            were added. So by default, the most recently added
            filesystem will be looked at first.
        :type priority: int

        """

        if isinstance(fs, text_type):
            fs = open_fs(fs)

        if not isinstance(fs, FS):
            raise TypeError(
                'fs argument should be an FS object or FS URL'
            )

        self._filesystems[name] = _PrioritizedFS(
            priority=(priority, self._sort_index),
            fs=fs
        )
        self._sort_index += 1
        self._resort()

        if write:
            self.write_fs = fs
            self._write_fs_name = name

    def get_fs(self, name):
        """
        Get a filesystem from its name.

        :param name: The name of a filesystem previously added.
        :type name: str

        """
        return self._filesystems[name].fs

    def _resort(self):
        """Force iterate_fs to re-sort on next reference."""
        self._fs_sequence = None

    def iterate_fs(self):
        """Get iterator that returns (name, fs) in priority order."""
        if self._fs_sequence is None:
            self._fs_sequence = [
                (name, fs)
                for name, (_order, fs) in
                sorted(
                    self._filesystems.items(),
                    key=itemgetter(1),
                    reverse=True
                )
            ]
        return iter(self._fs_sequence)

    def _delegate(self, path):
        """Get a filesystem which has a given path."""
        for _name, fs in self.iterate_fs():
            if fs.exists(path):
                return fs
        return None

    def _delegate_required(self, path):
        fs = self._delegate(path)
        if fs is None:
            raise errors.ResourceNotFound(path)
        return fs

    def _require_writable(self, path):
        if self.write_fs is None:
            raise errors.ResourceReadOnly(path)

    def which(self, path, mode="r"):
        """
        Get a tuple of (name, filesystem) that the given path would map
        to.

        :param path: A path on the filesystem.
        :type path: str
        :param mode: A open mode.
        :type mode: str

        """

        if check_writable(mode):
            return self._write_fs_name, self.write_fs
        for name, fs in self.iterate_fs():
            if fs.exists(path):
                return name, fs
        return None, None

    def close(self):
        self._closed = True
        if self._auto_close:
            try:
                for _order, fs in self._filesystems.values():
                    fs.close()
            finally:
                self._filesystems.clear()
                self._resort()

    def getinfo(self, path, namespaces=None):
        self.check()
        namespaces = namespaces or ()
        fs = self._delegate(path)
        if fs is None:
            raise errors.ResourceNotFound(path)
        _path = abspath(normpath(path))
        info = fs.getinfo(_path, namespaces=namespaces)
        return info

    def listdir(self, path):
        self.check()
        directory = []
        exists = False
        for _name, _fs in self.iterate_fs():
            try:
                directory.extend(_fs.listdir(path))
            except errors.ResourceNotFound:
                pass
            else:
                exists = True
        if not exists:
            raise errors.ResourceNotFound(path)
        return directory

    def makedir(self, path, permissions=None, recreate=False):
        self.check()
        self._require_writable(path)
        return self.write_fs.makedir(
            path, permissions=permissions, recreate=recreate)

    def openbin(self, path, mode='r', buffering=-1, **options):
        self.check()
        if check_writable(mode):
            self._require_writable(path)
            _fs = self.write_fs
        else:
            _fs = self._delegate_required(path)
        return _fs.openbin(
            path,
            mode=mode,
            buffering=buffering,
            **options
        )

    def remove(self, path):
        self.check()
        fs = self._delegate_required(path)
        return fs.remove(path)

    def removedir(self, path):
        self.check()
        fs = self._delegate_required(path)
        return fs.removedir(path)

    def scandir(self, path, namespaces=None, page=None):
        self.check()
        seen = set()
        exists = False
        for _name, fs in self.iterate_fs():
            try:
                for info in fs.scandir(path, namespaces=namespaces, page=page):
                    if info.name not in seen:
                        yield info
                        seen.add(info.name)
                exists = True
            except errors.ResourceNotFound:
                pass

        if not exists:
            raise errors.ResourceNotFound(path)

    def getbytes(self, path):
        self.check()
        fs = self._delegate(path)
        if fs is None:
            raise errors.ResourceNotFound(path)
        return fs.getbytes(path)

    def gettext(self, path, encoding=None, errors=None, newline=''):
        self.check()
        fs = self._delegate_required(path)
        return fs.gettext(
            path,
            encoding=encoding,
            errors=errors,
            newline=newline
        )

    def getsize(self, path):
        self.check()
        fs = self._delegate_required(path)
        return fs.getsize(path)

    def getsyspath(self, path):
        self.check()
        fs = self._delegate_required(path)
        return fs.getsyspath(path)

    def gettype(self, path):
        self.check()
        fs = self._delegate_required(path)
        return fs.gettype(path)

    def geturl(self, path, purpose='download'):
        self.check()
        fs = self._delegate_required(path)
        return fs.geturl(path, purpose=purpose)

    def hassyspath(self, path):
        self.check()
        fs = self._delegate_required(path)
        return fs.hassyspath(path)

    def hasurl(self, path, purpose='download'):
        self.check()
        fs = self._delegate_required(path)
        return fs.hasurl(path, purpose=purpose)

    def isdir(self, path):
        self.check()
        fs = self._delegate(path)
        return fs and fs.isdir(path)

    def isfile(self, path):
        self.check()
        fs = self._delegate(path)
        return fs and fs.isfile(path)

    def setinfo(self, path, info):
        self.check()
        self._require_writable(path)
        return self.write_fs.setinfo(path, info)

    def validatepath(self, path):
        self.check()
        if self.write_fs is not None:
            self.write_fs.validatepath(path)
        else:
            super(MultiFS, self).validatepath(path)

    def makedirs(self, path, permissions=None, recreate=False):
        self.check()
        self._require_writable(path)
        return self.write_fs.makedirs(
            path,
            permissions=permissions,
            recreate=recreate
        )

    def open(self,
             path,
             mode='r',
             buffering=-1,
             encoding=None,
             errors=None,
             newline='',
             **kwargs):
        self.check()
        if check_writable(mode):
            self._require_writable(path)
            _fs = self.write_fs
        else:
            _fs = self._delegate_required(path)
        return _fs.open(
            path,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            **kwargs
        )

    def setbinfile(self, path, file):
        self._require_writable(path)
        self.write_fs.setbinfile(path, file)

    def setbytes(self, path, contents):
        self._require_writable(path)
        return self.write_fs.setbytes(path, contents)

    def settext(self,
                path,
                contents,
                encoding='utf-8',
                errors=None,
                newline=''):
        self._require_writable(path)
        return self.write_fs.settext(
            path,
            contents,
            encoding=encoding,
            errors=errors,
            newline=newline
        )
